------------------------------------------------------------------
--GridViewController.lua
--Author: ingeniousclown

--This is mostly a re-implementation of ZO_ScrollList_UpdateScroll
--and copies of all the local functions necessary to make it work.

--It also includes an init to add all the necessary fields to
--convert the list to grid view.
------------------------------------------------------------------

local GRID_SLOT_BACKGROUND = "BankGridView/assets/gridItem_outline.dds"
local LIST_SLOT_BACKGROUND = [[EsoUI/Art/Miscellaneous/listItem_backdrop.dds]]

-------------------------
--my own util functions
-------------------------

local function GetTopLeftTargetPosition(index, itemsPerRow, controlWidth, controlHeight, leftPadding, gridSpacing)
    controlTop = zo_floor((index-1) / itemsPerRow) * (controlHeight + gridSpacing)
    controlLeft = ((index-1) % itemsPerRow) * (controlWidth + gridSpacing) + leftPadding

    return controlTop, controlLeft
end

local function AddColor(control)
    local bagId = control.dataEntry.data.bagId
    local slotIndex = control.dataEntry.data.slotIndex
    local _, _, _, _, _, _, _, quality = GetItemInfo(bagId, slotIndex)
    local r, g, b = GetInterfaceColor(INTERFACE_COLOR_TYPE_ITEM_QUALITY_COLORS, quality)

    control:GetNamedChild("Bg"):SetColor(r, g, b, 1)
end

local function ReshapeSlot(control, isGrid, width, forceUpdate)
    local backpack = ZO_PlayerBankBackpack
    local thisControl = control

    if control == nil then return end

    if(control.isGrid == nil) then control.isGrid = false end

    --i pulled this out of the below part of the function because it wasn't
    --being hidden correctly.  My guess is that some other function unhides
    --the control
    control:GetNamedChild("SellPrice"):SetHidden(isGrid)
    if(control.isGrid and control.dataEntry and backpack.showGridOutlines) then
        AddColor(control)
    end

    if((control.isGrid ~= isGrid or (isGrid and forceUpdate)) and control.dataEntry) then
        control.isGrid = isGrid
        local thisName = control:GetName()

        local button = control:GetNamedChild("Button")
        local bg = control:GetNamedChild("Bg")
        local new = control:GetNamedChild("NewStatus")
        local name = control:GetNamedChild("Name")
        local stat = control:GetNamedChild("StatValue")
        -- local sell = control:GetNamedChild("SellPrice")
        local highlight = control:GetNamedChild("Highlight")

        button:ClearAnchors()
        new:ClearAnchors()
        control:SetWidth(width)
        if(isGrid) then
            button:SetAnchor(CENTER, control, CENTER)

            new:SetDimensions(5,5)
            new:SetAnchor(TOPLEFT, control, TOPLEFT, 10, 10)

            name:SetHidden(true)
            stat:SetHidden(true)

            if(backpack.showGridOutlines) then
                bg:SetTexture(GRID_SLOT_BACKGROUND)
                bg:SetTextureCoords(0, 1, 0, 1)
            else
                bg:SetTexture(LIST_SLOT_BACKGROUND)
                bg:SetTextureCoords(0, 1, 0, .8125)
                bg:SetColor(0, 0, 0, 1)
            end

            --highlight:SetTexture(GRID_SLOT_HIGHLIGHT)
        else
            button:SetAnchor(CENTER, control, TOPLEFT, 47, 26)

            new:SetAnchor(CENTER, control, TOPLEFT, 27, 27)

            name:SetHidden(false)
            stat:SetHidden(false)

            bg:SetTexture(LIST_SLOT_BACKGROUND)
            bg:SetTextureCoords(0, 1, 0, .8125)
            bg:SetColor(0, 0, 0, 1)
        end
    end
end

local function ReshapeSlots(self)
    local allControlsParent = self:GetNamedChild("Contents")
    local numControls = allControlsParent:GetNumChildren()
    
    local width
    if self.isGrid then width = self.controlWidth else width = self.contentsWidth end

    for i = 2, numControls do
        ReshapeSlot(allControlsParent:GetChild(i), self.isGrid, width, self.forceUpdate)
    end
end

--------------------------------------------------------------------------------------
--the following functions are ported from scrolltemplates.lua because they're local
--and necessary for the UpdateList function to work correctly
--------------------------------------------------------------------------------------

--scrolltemplates.lua line 139
local function UpdateScrollFade(useFadeGradient, scroll, slider, sliderValue)
    if(useFadeGradient) then
        local sliderMin, sliderMax = slider:GetMinMax()
        sliderValue = sliderValue or slider:GetValue()

        if(sliderValue > sliderMin) then
            scroll:SetFadeGradient(1, 0, 1, zo_min(sliderValue - sliderMin, 64))
        else
            scroll:SetFadeGradient(1, 0, 0, 0)
        end
        
        if(sliderValue < sliderMax) then
            scroll:SetFadeGradient(2, 0, -1, zo_min(sliderMax - sliderValue, 64))
        else
            scroll:SetFadeGradient(2, 0, 0, 0);
        end
    else
        scroll:SetFadeGradient(1, 0, 0, 0)
        scroll:SetFadeGradient(2, 0, 0, 0)
    end
end

--scrolltemplates.lua line 425
local function AreSelectionsEnabled(self)
    return self.selectionTemplate or self.selectionCallback
end

--scrolltemplates.lua line 580
local function RemoveAnimationOnControl(control, animationFieldName)
    if control[animationFieldName] then
        control[animationFieldName]:PlayBackward()
    end
end

--scrolltemplates.lua line 596
local function UnhighlightControl(self, control) 
    RemoveAnimationOnControl(control, "HighlightAnimation")

    self.highlightedControl = nil

    if(self.highlightCallback) then
        self.highlightCallback(control, false)
    end
end

--scrolltemplates.lua line 612
local function UnselectControl(self, control)
    RemoveAnimationOnControl(control, "SelectionAnimation")

    self.selectedControl = nil
end

--scrolltemplates.lua line 740
--Determines if one piece of selected data is the "same" as the other. Used mainly to
--keep an item selected even when the data for the list is updated if they share some
--property determined by the equality function. For example, if you have an item with
--id=1 and state=up and replace it with id=1 and state=down, the selection will be maintained
--if the equality function only compares ids.  
local function AreDataEqualSelections(self, data1, data2)
    if(data1 == data2) then
        return true
    end

    if(data1 == nil or data2 == nil) then
        return false
    end        

    local dataEntry1 = data1.dataEntry
    local dataEntry2 = data2.dataEntry
    if(dataEntry1.typeId == dataEntry2.typeId) then
        local equalityFunction = self.dataTypes[dataEntry1.typeId].equalityFunction
        if(equalityFunction) then
            return equalityFunction(data1, data2)
        end
    end

    return false
end

--scrolltemplates.lua line 825
local function FreeActiveScrollListControl(self, i)
    local currentControl = self.activeControls[i]
    local currentDataEntry = currentControl.dataEntry
    local dataType = self.dataTypes[currentDataEntry.typeId]
    
    if(self.highlightTemplate and currentControl == self.highlightedControl) then
        UnhighlightControl(self, currentControl)
        if(self.highlightLocked) then
            self.highlightLocked = false
        end
    end

    if(currentControl == self.pendingHighlightControl) then
        self.pendingHighlightControl = nil
    end
    
    if(AreSelectionsEnabled(self) and currentControl == self.selectedControl) then
        UnselectControl(self, currentControl)
    end
    
    if(dataType.hideCallback) then
        dataType.hideCallback(currentControl, currentControl.dataEntry.data)
    end
    
    dataType.pool:ReleaseObject(currentControl.key)
    currentControl.key = nil
    currentControl.dataEntry = nil
    self.activeControls[i] = self.activeControls[#self.activeControls]
    self.activeControls[#self.activeControls] = nil
end
--end copied functions

--------------------------------------------------------------------------------------
--meat of the mod - modified ZO_ScrollList_UpdateScroll(self)
--------------------------------------------------------------------------------------

--modified version of function ZO_ScrollList_UpdateScroll(self)
--this is lightly modified to turn the view into a grid view
--from esoui\libraries\zo_templates\scrolltemplates.lua line 1157
local consideredMap = {}

local function UpdateScroll_Grid(self)
    local windowHeight = ZO_ScrollList_GetHeight(self)
    local controlHeight = self.controlHeight
    local controlWidth = self.controlWidth
    local activeControls = self.activeControls
    local leftPadding = self.leftPadding
    local gridSpacing = self.gridSpacing
    local offset = self.offset

    UpdateScrollFade(self.useFadeGradient, self.contents, self.scrollbar, offset)
    
    --remove active controls that are now hidden
    local i = 1
    local numActive = #activeControls
    while(i <= numActive) do
        local currentDataEntry = activeControls[i].dataEntry
        
        if(currentDataEntry.bottom < offset or currentDataEntry.top > offset + windowHeight) then
            FreeActiveScrollListControl(self, i)
            numActive = numActive - 1
        else
            i = i + 1
        end
        
        consideredMap[currentDataEntry] = true
    end
        
    --add revealed controls
    local firstInViewIndex = zo_floor(offset / controlHeight)+1
   
    local data = self.data
    local dataTypes = self.dataTypes
    local visibleData = self.visibleData
    local mode = self.mode
    
    local i = firstInViewIndex
    local visibleDataIndex = visibleData[i]
    local dataEntry = data[visibleDataIndex]
    local bottomEdge = offset + windowHeight

    local itemsPerRow = zo_floor((ZO_PlayerBankBackpackContents:GetWidth() - leftPadding) / (controlWidth + gridSpacing))
    
    local controlTop, controlLeft
    
    --removed isUniform check because we're assuming always uniform
    if(dataEntry) then
        controlTop, controlLeft = GetTopLeftTargetPosition(i, itemsPerRow, controlWidth, controlHeight, leftPadding, gridSpacing)
    end
    while(dataEntry and controlTop <= bottomEdge) do
        if(not consideredMap[dataEntry]) then
            local dataType = dataTypes[dataEntry.typeId]
            local controlPool = dataType.pool
            local control, key = controlPool:AcquireObject()
            
            control:SetHidden(false)
            control.dataEntry = dataEntry
            control.key = key
            if(dataType.setupCallback) then
                dataType.setupCallback(control, dataEntry.data, self)
            end
            table.insert(activeControls, control)
            consideredMap[dataEntry] = true
            
            if(AreDataEqualSelections(self, dataEntry.data, self.selectedData)) then
                SelectControl(self, control)
            end
            
            --even uniform active controls need to know their position to determine if they are still active

            dataEntry.top = controlTop
            dataEntry.bottom = controlTop + controlHeight
            dataEntry.left = controlLeft
            dataEntry.right = controlLeft + controlWidth
        end
        i = i + 1
        visibleDataIndex = visibleData[i]
        dataEntry = data[visibleDataIndex]
        --removed isUniform check because we're assuming always uniform
        if(dataEntry) then
            controlTop, controlLeft = GetTopLeftTargetPosition(i, itemsPerRow, controlWidth, controlHeight, leftPadding, gridSpacing)
        end
    end
    
    --update positions
    local contents = self.contents
    local numActive = #activeControls
    
    for i = 1, numActive do
        local currentControl = activeControls[i]
        local currentData = currentControl.dataEntry
        local controlOffset = currentData.top - offset
        local controlOffsetX = currentData.left

        currentControl:ClearAnchors()
        currentControl:SetAnchor(TOPLEFT, contents, TOPLEFT, controlOffsetX, controlOffset)
        --removed other anchor because this will no longer stretch across the contents pane
    end  
    
    --reset considered
    for k,v in pairs(consideredMap) do
        consideredMap[k] = nil
    end

    --reshape the slots to finish the illusion
    --this is done in the toggle function, but this ensures that any NEW slots get reshaped
    ReshapeSlots(self)
end

--interface to toggle grid view on and off
function BankGridView_ToggleGrid(toggle)
    local self = ZO_PlayerBankBackpack
    self.isGrid = toggle    
    ZO_ScrollList_ResetToTop(self)

    while(#self.activeControls > 0) do
        FreeActiveScrollListControl(self, 1)
    end

    if(toggle) then
        UpdateScroll_Grid(self)
    else
        ZO_ScrollList_UpdateScroll(self)
    end

    ReshapeSlots(self)
    return
end

function BankGridView_ToggleOutlines(toggle)
    local self = ZO_PlayerBankBackpack
    self.showGridOutlines = toggle
    self.forceUpdate = (toggle and self.isGrid)

    --no need to update if current in list view
    if not self.forceUpdate then return end

    while(#self.activeControls > 0) do
        FreeActiveScrollListControl(self, 1)
    end

    if(toggle) then
        UpdateScroll_Grid(self)
    else
        ZO_ScrollList_UpdateScroll(self)
    end

    ReshapeSlots(self)
end

--hook function!  to be called before "ZO_ScrollList_UpdateScroll"
--thanks to Seerah for teaching me about this possibility
local function ScrollController(self)
    if(self.isGrid) then
        UpdateScroll_Grid(self)
        return true
    else
        return false
    end
end

--add necessary fields to the default UI's controls to facilitate the grid view
function InitGridView()
    local controlWidth = ZO_PlayerBankBackpack.controlHeight
    local leftPadding = 25
    local contentsWidth = ZO_PlayerBankBackpack:GetNamedChild("Contents"):GetWidth()
    local itemsPerRow = zo_floor((contentsWidth - leftPadding) / (controlWidth))
    local gridSpacing = ((contentsWidth - leftPadding) % itemsPerRow) / itemsPerRow

    ZO_PlayerBankBackpack.isGrid = true
    ZO_PlayerBankBackpack.showGridOutlines = true
    ZO_PlayerBankBackpack.forceUpdate = false
    ZO_PlayerBankBackpack.controlWidth = controlWidth
    ZO_PlayerBankBackpack.leftPadding = leftPadding
    ZO_PlayerBankBackpack.contentsWidth = contentsWidth
    ZO_PlayerBankBackpack.itemsPerRow = itemsPerRow
    ZO_PlayerBankBackpack.gridSpacing = gridSpacing

    ZO_PreHook("ZO_ScrollList_UpdateScroll", ScrollController)
end